@@ -144,6 +144,7 @@ enum {
 enum
 {
     SIG_CLOSED = 0,
+    SIG_NEED_POSITION,
     SIG_ACTION_INVOKED,
     N_SIGS,
 };
@@ -162,6 +163,8 @@ static void xfce_notify_window_get_property(GObject *object,
 static void xfce_notify_window_notify(GObject *object,
                                       GParamSpec *pspec);
 
+static void xfce_notify_window_size_allocate(GtkWidget *widget,
+                                             GtkAllocation *allocation);
 static void xfce_notify_window_realize(GtkWidget *widget);
 static void xfce_notify_window_show(GtkWidget *widget);
 static void xfce_notify_window_unrealize(GtkWidget *widget);
@@ -211,6 +214,8 @@ static void xfce_notify_window_set_do_slideout(XfceNotifyWindow *window,
 
 static void xfce_notify_window_move(XfceNotifyWindow *window, gint x, gint y);
 
+static void xfce_notify_window_ensure_widgets(XfceNotifyWindow *window);
+
 
 static guint signals[N_SIGS] = { 0, };
 
@@ -231,6 +236,7 @@ xfce_notify_window_class_init(XfceNotifyWindowClass *klass)
     gobject_class->get_property = xfce_notify_window_get_property;
     gobject_class->notify = xfce_notify_window_notify;
 
+    widget_class->size_allocate = xfce_notify_window_size_allocate;
     widget_class->realize = xfce_notify_window_realize;
     widget_class->unrealize = xfce_notify_window_unrealize;
     widget_class->show = xfce_notify_window_show;
@@ -369,6 +375,16 @@ xfce_notify_window_class_init(XfceNotifyWindowClass *klass)
                                        G_TYPE_NONE, 1,
                                        XFCE_TYPE_NOTIFY_CLOSE_REASON);
 
+    signals[SIG_NEED_POSITION] = g_signal_new("need-position",
+                                              XFCE_TYPE_NOTIFY_WINDOW,
+                                              G_SIGNAL_RUN_LAST,
+                                              0,
+                                              NULL,
+                                              NULL,
+                                              g_cclosure_marshal_VOID__VOID,
+                                              G_TYPE_NONE,
+                                              0);
+
     signals[SIG_ACTION_INVOKED] = g_signal_new("action-invoked",
                                                XFCE_TYPE_NOTIFY_WINDOW,
                                                G_SIGNAL_RUN_LAST,
@@ -403,16 +419,6 @@ xfce_notify_window_init(XfceNotifyWindow *window)
     window->original_y = G_MININT;
     window->op_change_steps = FADE_TIME / FADE_CHANGE_TIMEOUT;
 
-#ifdef ENABLE_WAYLAND
-    if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
-        gtk_layer_init_for_window(GTK_WINDOW(window));
-        gtk_layer_set_layer(GTK_WINDOW(window), GTK_LAYER_SHELL_LAYER_OVERLAY);
-        gtk_layer_set_namespace(GTK_WINDOW(window), "notification");
-        gtk_layer_set_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_TOP, TRUE);
-        gtk_layer_set_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_LEFT, TRUE);
-    }
-#endif
-
     gtk_widget_set_name (GTK_WIDGET(window), "XfceNotifyWindow");
     gtk_window_set_keep_above(GTK_WINDOW(window), TRUE);
     gtk_window_stick(GTK_WINDOW(window));
@@ -443,80 +449,57 @@ xfce_notify_window_init(XfceNotifyWindow *window)
 static void
 xfce_notify_window_constructed(GObject *object) {
     XfceNotifyWindow *window = XFCE_NOTIFY_WINDOW(object);
-    GtkWidget *topvbox, *tophbox, *vbox;
-    gint screen_width;
-    gdouble padding = DEFAULT_PADDING;
     GtkCssProvider *provider;
-    GdkRectangle geometry;
 
     G_OBJECT_CLASS(xfce_notify_window_parent_class)->constructed(object);
 
-    /* Use the monitor width to get a maximum width for the notification bubble.
-       This assumes that a character is 10px wide and we want a third of the
-       monitor as maximum width. */
-    gdk_monitor_get_geometry(window->monitor, &geometry);
-    screen_width = geometry.width / 30;
+#ifdef ENABLE_WAYLAND
+    if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
+        gtk_layer_init_for_window(GTK_WINDOW(window));
+        if (window->monitor != NULL) {
+            gtk_layer_set_monitor(GTK_WINDOW(window), window->monitor);
+        }
+        gtk_layer_set_layer(GTK_WINDOW(window), GTK_LAYER_SHELL_LAYER_OVERLAY);
+        gtk_layer_set_namespace(GTK_WINDOW(window), "notification");
 
-    gtk_widget_style_get(GTK_WIDGET(window),
-                         "padding", &padding,
-                         NULL);
+        switch (window->notify_location) {
+            case XFCE_NOTIFY_POS_TOP_LEFT:
+            case XFCE_NOTIFY_POS_TOP_RIGHT:
+            case XFCE_NOTIFY_POS_TOP_CENTER:
+                gtk_layer_set_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_TOP, TRUE);
+                break;
+            case XFCE_NOTIFY_POS_BOTTOM_LEFT:
+            case XFCE_NOTIFY_POS_BOTTOM_RIGHT:
+            case XFCE_NOTIFY_POS_BOTTOM_CENTER:
+                gtk_layer_set_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_BOTTOM, TRUE);
+                break;
+            default:
+                g_assert_not_reached();
+                break;
+        }
 
-    topvbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
-    gtk_box_set_homogeneous (GTK_BOX (topvbox), FALSE);
-    gtk_container_set_border_width (GTK_CONTAINER(topvbox), padding);
-    gtk_widget_show (topvbox);
-    gtk_container_add (GTK_CONTAINER(window), topvbox);
-    tophbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
-    gtk_box_set_homogeneous (GTK_BOX (tophbox), FALSE);
-    gtk_widget_show (tophbox);
-    gtk_container_add (GTK_CONTAINER(topvbox), tophbox);
+        switch (window->notify_location) {
+            case XFCE_NOTIFY_POS_TOP_LEFT:
+            case XFCE_NOTIFY_POS_BOTTOM_LEFT:
+                gtk_layer_set_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_LEFT, TRUE);
+                break;
+            case XFCE_NOTIFY_POS_TOP_RIGHT:
+            case XFCE_NOTIFY_POS_BOTTOM_RIGHT:
+                gtk_layer_set_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_RIGHT, TRUE);
+                break;
+            case XFCE_NOTIFY_POS_TOP_CENTER:
+            case XFCE_NOTIFY_POS_BOTTOM_CENTER:
+                // We're not quite sure which side we're going to anchor to
+                // yet, but we'll decide when we're told our position.
+                break;
+            default:
+                g_assert_not_reached();
+                break;
+        }
+    }
+#endif
 
-    window->icon_box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
-    gtk_container_set_border_width(GTK_CONTAINER(window->icon_box), 0);
-    gtk_widget_set_margin_end (GTK_WIDGET (window->icon_box), padding);
-
-    gtk_box_pack_start(GTK_BOX(tophbox), window->icon_box, FALSE, TRUE, 0);
-
-    window->icon = gtk_image_new();
-    gtk_widget_show(window->icon);
-    gtk_container_add(GTK_CONTAINER(window->icon_box), window->icon);
-
-    window->content_box = vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
-    gtk_box_set_homogeneous(GTK_BOX (vbox), FALSE);
-    gtk_container_set_border_width(GTK_CONTAINER(vbox), 0);
-    gtk_widget_show(vbox);
-    gtk_box_pack_start(GTK_BOX(tophbox), vbox, TRUE, TRUE, 0);
-
-    window->summary = gtk_label_new(NULL);
-    gtk_widget_set_name (window->summary, "summary");
-    gtk_label_set_max_width_chars (GTK_LABEL(window->summary), screen_width);
-    gtk_label_set_ellipsize (GTK_LABEL (window->summary), PANGO_ELLIPSIZE_END);
-    gtk_label_set_line_wrap (GTK_LABEL(window->summary), TRUE);
-    gtk_label_set_line_wrap_mode (GTK_LABEL (window->summary), PANGO_WRAP_WORD_CHAR);
-    gtk_label_set_lines (GTK_LABEL (window->summary), 1);
-    gtk_widget_set_halign (window->summary, GTK_ALIGN_FILL);
-    gtk_label_set_xalign (GTK_LABEL(window->summary), 0);
-    gtk_widget_set_valign (window->summary, GTK_ALIGN_BASELINE);
-    gtk_box_pack_start(GTK_BOX(vbox), window->summary, TRUE, TRUE, 0);
-
-    window->body = gtk_label_new(NULL);
-    gtk_widget_set_name (window->body, "body");
-    gtk_label_set_max_width_chars (GTK_LABEL(window->body), screen_width);
-    gtk_label_set_ellipsize (GTK_LABEL (window->body), PANGO_ELLIPSIZE_END);
-    gtk_label_set_line_wrap (GTK_LABEL(window->body), TRUE);
-    gtk_label_set_line_wrap_mode (GTK_LABEL (window->body), PANGO_WRAP_WORD_CHAR);
-    gtk_label_set_lines (GTK_LABEL (window->body), 6);
-    gtk_widget_set_halign (window->body, GTK_ALIGN_FILL);
-    gtk_label_set_xalign (GTK_LABEL(window->body), 0);
-    gtk_widget_set_valign (window->body, GTK_ALIGN_BASELINE);
-    gtk_box_pack_start(GTK_BOX(vbox), window->body, TRUE, TRUE, 0);
-
-    window->button_box = gtk_button_box_new(GTK_ORIENTATION_HORIZONTAL);
-    gtk_button_box_set_layout(GTK_BUTTON_BOX(window->button_box),
-                              GTK_BUTTONBOX_END);
-    gtk_box_set_spacing (GTK_BOX(window->button_box), padding / 2);
-    gtk_box_set_homogeneous (GTK_BOX(window->button_box), FALSE);
-    gtk_box_pack_end (GTK_BOX(topvbox), window->button_box, FALSE, FALSE, 0);
+    xfce_notify_window_ensure_widgets(window);
 
     gtk_style_context_add_class (gtk_widget_get_style_context (GTK_WIDGET (window)), "xfce4-notifyd");
     provider = gtk_css_provider_new ();
@@ -577,7 +560,9 @@ xfce_notify_window_finalize(GObject *object)
 {
     XfceNotifyWindow *window = XFCE_NOTIFY_WINDOW(object);
 
-    g_object_unref(window->monitor);
+    if (window->monitor != NULL) {
+        g_object_unref(window->monitor);
+    }
 
     if (window->css_provider != NULL) {
         g_object_unref(window->css_provider);
@@ -777,13 +762,23 @@ xfce_notify_window_notify(GObject *object, GParamSpec *pspec) {
     }
 }
 
+static void
+xfce_notify_window_size_allocate(GtkWidget *widget, GtkAllocation *allocation) {
+    GTK_WIDGET_CLASS(xfce_notify_window_parent_class)->size_allocate(widget, allocation);
+
+    XfceNotifyWindow *window = XFCE_NOTIFY_WINDOW(widget);
+    if (window->geometry.width != allocation->width || window->geometry.height != allocation->height) {
+        g_signal_emit(widget, signals[SIG_NEED_POSITION], 0);
+    }
+}
+
 static void
 xfce_notify_window_realize(GtkWidget *widget)
 {
     XfceNotifyWindow *window = XFCE_NOTIFY_WINDOW(widget);
 
 #ifdef ENABLE_WAYLAND
-    if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
+    if (window->monitor != NULL && GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
         gtk_layer_set_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_LEFT, window->geometry.x);
         gtk_layer_set_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_TOP, window->geometry.y);
     }
@@ -840,7 +835,13 @@ xfce_notify_window_show(GtkWidget *widget) {
 
     GTK_WIDGET_CLASS(xfce_notify_window_parent_class)->show(widget);
 
-    xfce_notify_window_start_expiration(window);
+    if (window->monitor != NULL) {
+        xfce_notify_window_start_expiration(window);
+    } else {
+        // If we don't know the monitor yet, we need to 'hide' the window until
+        // we figure out what monitor it's on.
+        gtk_widget_set_opacity(widget, 0.0);
+    }
 }
 
 static void
@@ -974,6 +975,30 @@ xfce_notify_window_configure_event(GtkWidget *widget,
     return ret;
 }
 
+#ifdef ENABLE_WAYLAND
+static gint
+wayland_get_x_margin(XfceNotifyWindow *window) {
+    if (gtk_layer_get_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_LEFT)) {
+        return gtk_layer_get_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_LEFT);
+    } else if (gtk_layer_get_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_RIGHT)) {
+        return gtk_layer_get_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_RIGHT);
+    } else {
+        g_assert_not_reached();
+    }
+}
+
+static gint
+wayland_get_y_margin(XfceNotifyWindow *window) {
+    if (gtk_layer_get_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_TOP)) {
+        return gtk_layer_get_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_TOP);
+    } else if (gtk_layer_get_anchor(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_BOTTOM)) {
+        return gtk_layer_get_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_BOTTOM);
+    } else {
+        g_assert_not_reached();
+    }
+}
+#endif
+
 static gboolean
 xfce_notify_window_expire_timeout(gpointer data)
 {
@@ -994,8 +1019,8 @@ xfce_notify_window_expire_timeout(gpointer data)
             if (window->do_slideout) {
 #ifdef ENABLE_WAYLAND
                 if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
-                    window->original_x = gtk_layer_get_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_LEFT);
-                    window->original_y = gtk_layer_get_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_TOP);
+                    window->original_x = wayland_get_x_margin(window);
+                    window->original_y = wayland_get_y_margin(window);
                 } else
 #endif
                 {
@@ -1038,14 +1063,22 @@ xfce_notify_window_fade_timeout(gpointer data)
              (window->notify_location == XFCE_NOTIFY_POS_TOP_CENTER ||
               window->notify_location == XFCE_NOTIFY_POS_BOTTOM_CENTER));
 
+#ifdef ENABLE_WAYLAND
+        if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
+            // On Wayland we are always considering a margin, or a distance
+            // from the edge of the screen, so we always subtract pixels.
+            add_pixels = FALSE;
+        }
+#endif
+
         gdk_monitor_get_geometry(window->monitor, &monitor_geom);
 
         if (window->draw_offset_x == 0 && window->draw_offset_y == 0) {
 
 #ifdef ENABLE_WAYLAND
             if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
-                x = gtk_layer_get_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_LEFT);
-                y = gtk_layer_get_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_TOP);
+                x = wayland_get_x_margin(window);
+                y = wayland_get_y_margin(window);
             } else
 #endif
             {
@@ -1563,20 +1596,108 @@ xfce_notify_window_set_do_slideout(XfceNotifyWindow *window, gboolean do_slideou
 
 static void
 xfce_notify_window_move(XfceNotifyWindow *window, gint x, gint y) {
+    DBG("entering, (%d, %d)", x, y);
 
 #ifdef ENABLE_X11
     if (GDK_IS_X11_DISPLAY(gdk_display_get_default())) {
         gtk_window_move(GTK_WINDOW(window), x, y);
     }
 #endif
+
 #ifdef ENABLE_WAYLAND
     if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
+        // Only two anchors are actually set, one for x and one for y, but it
+        // doesn't hurt to set them all, and this way we don't have to check.
         gtk_layer_set_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_LEFT, x);
+        gtk_layer_set_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_RIGHT, x);
         gtk_layer_set_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_TOP, y);
+        gtk_layer_set_margin(GTK_WINDOW(window), GTK_LAYER_SHELL_EDGE_BOTTOM, y);
     }
 #endif
 }
 
+static void
+xfce_notify_window_ensure_widgets(XfceNotifyWindow *window) {
+    if (window->icon == NULL) {
+        GtkWidget *topvbox, *tophbox, *vbox;
+        gdouble padding = DEFAULT_PADDING;
+
+        gtk_widget_style_get(GTK_WIDGET(window),
+                             "padding", &padding,
+                             NULL);
+
+        topvbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
+        gtk_box_set_homogeneous (GTK_BOX (topvbox), FALSE);
+        gtk_container_set_border_width (GTK_CONTAINER(topvbox), padding);
+        gtk_widget_show (topvbox);
+        gtk_container_add (GTK_CONTAINER(window), topvbox);
+        tophbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
+        gtk_box_set_homogeneous (GTK_BOX (tophbox), FALSE);
+        gtk_widget_show (tophbox);
+        gtk_container_add (GTK_CONTAINER(topvbox), tophbox);
+
+        window->icon_box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
+        gtk_container_set_border_width(GTK_CONTAINER(window->icon_box), 0);
+        gtk_widget_set_margin_end (GTK_WIDGET (window->icon_box), padding);
+
+        gtk_box_pack_start(GTK_BOX(tophbox), window->icon_box, FALSE, TRUE, 0);
+
+        window->icon = gtk_image_new();
+        gtk_widget_show(window->icon);
+        gtk_container_add(GTK_CONTAINER(window->icon_box), window->icon);
+
+        window->content_box = vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
+        gtk_box_set_homogeneous(GTK_BOX (vbox), FALSE);
+        gtk_container_set_border_width(GTK_CONTAINER(vbox), 0);
+        gtk_widget_show(vbox);
+        gtk_box_pack_start(GTK_BOX(tophbox), vbox, TRUE, TRUE, 0);
+
+        window->summary = gtk_label_new(NULL);
+        gtk_widget_set_name (window->summary, "summary");
+        gtk_label_set_ellipsize (GTK_LABEL (window->summary), PANGO_ELLIPSIZE_END);
+        gtk_label_set_line_wrap (GTK_LABEL(window->summary), TRUE);
+        gtk_label_set_line_wrap_mode (GTK_LABEL (window->summary), PANGO_WRAP_WORD_CHAR);
+        gtk_label_set_lines (GTK_LABEL (window->summary), 1);
+        gtk_widget_set_halign (window->summary, GTK_ALIGN_FILL);
+        gtk_label_set_xalign (GTK_LABEL(window->summary), 0);
+        gtk_widget_set_valign (window->summary, GTK_ALIGN_BASELINE);
+        gtk_box_pack_start(GTK_BOX(vbox), window->summary, TRUE, TRUE, 0);
+
+        window->body = gtk_label_new(NULL);
+        gtk_widget_set_name (window->body, "body");
+        gtk_label_set_ellipsize (GTK_LABEL (window->body), PANGO_ELLIPSIZE_END);
+        gtk_label_set_line_wrap (GTK_LABEL(window->body), TRUE);
+        gtk_label_set_line_wrap_mode (GTK_LABEL (window->body), PANGO_WRAP_WORD_CHAR);
+        gtk_label_set_lines (GTK_LABEL (window->body), 6);
+        gtk_widget_set_halign (window->body, GTK_ALIGN_FILL);
+        gtk_label_set_xalign (GTK_LABEL(window->body), 0);
+        gtk_widget_set_valign (window->body, GTK_ALIGN_BASELINE);
+        gtk_box_pack_start(GTK_BOX(vbox), window->body, TRUE, TRUE, 0);
+
+        window->button_box = gtk_button_box_new(GTK_ORIENTATION_HORIZONTAL);
+        gtk_button_box_set_layout(GTK_BUTTON_BOX(window->button_box),
+                                  GTK_BUTTONBOX_END);
+        gtk_box_set_spacing (GTK_BOX(window->button_box), padding / 2);
+        gtk_box_set_homogeneous (GTK_BOX(window->button_box), FALSE);
+        gtk_box_pack_end (GTK_BOX(topvbox), window->button_box, FALSE, FALSE, 0);
+    }
+
+    if (window->monitor != NULL) {
+        GdkRectangle geometry;
+        gint screen_width;
+
+        /* Use the monitor width to get a maximum width for the notification bubble.
+           This assumes that a character is 10px wide and we want a third of the
+           monitor as maximum width. */
+        gdk_monitor_get_geometry(window->monitor, &geometry);
+        screen_width = geometry.width / 30;
+
+        gtk_label_set_max_width_chars(GTK_LABEL(window->summary), screen_width);
+        gtk_label_set_max_width_chars(GTK_LABEL(window->body), screen_width);
+    }
+}
+
+
 GtkWidget *
 xfce_notify_window_new(guint id,
                        GdkMonitor *monitor,
@@ -1585,7 +1706,6 @@ xfce_notify_window_new(guint id,
                        gdouble normal_opacity,
                        gboolean show_text_with_gauge)
 {
-    g_return_val_if_fail(GDK_IS_MONITOR(monitor), NULL);
     return g_object_new(XFCE_TYPE_NOTIFY_WINDOW,
                         "type", GTK_WINDOW_TOPLEVEL,
                         "id", id,
@@ -1610,10 +1730,142 @@ xfce_notify_window_get_monitor(XfceNotifyWindow *window) {
 }
 
 void
-xfce_notify_window_set_geometry(XfceNotifyWindow *window, GdkRectangle rectangle) {
+xfce_notify_window_update_monitor(XfceNotifyWindow *window, GdkMonitor *monitor) {
     g_return_if_fail(XFCE_IS_NOTIFY_WINDOW(window));
-    window->geometry = rectangle;
-    xfce_notify_window_move(window, window->geometry.x, window->geometry.y);
+    g_return_if_fail(GDK_IS_MONITOR(monitor));
+    g_return_if_fail(window->monitor == NULL);
+
+    window->monitor = g_object_ref(monitor);
+    xfce_notify_window_ensure_widgets(window);
+    g_object_notify(G_OBJECT(window), "monitor");
+
+    xfce_notify_window_start_expiration(window);
+}
+
+void
+xfce_notify_window_set_geometry(XfceNotifyWindow *window, GdkRectangle *rectangle, GdkRectangle *monitor_workarea) {
+    g_return_if_fail(XFCE_IS_NOTIFY_WINDOW(window));
+
+    gint pos_x, pos_y;
+
+#ifdef ENABLE_X11
+    if (GDK_IS_X11_DISPLAY(gdk_display_get_default())) {
+        switch (window->notify_location) {
+            case XFCE_NOTIFY_POS_TOP_LEFT:
+                pos_x = monitor_workarea->x + rectangle->x;
+                pos_y = monitor_workarea->y + rectangle->y;
+                break;
+
+            case XFCE_NOTIFY_POS_TOP_RIGHT:
+                pos_x = monitor_workarea->x + monitor_workarea->width - rectangle->x - rectangle->width;
+                pos_y = monitor_workarea->y + rectangle->y;
+                break;
+
+            case XFCE_NOTIFY_POS_BOTTOM_LEFT:
+                pos_x = monitor_workarea->x + rectangle->x;
+                pos_y = monitor_workarea->y + monitor_workarea->height - rectangle->y - rectangle->height;
+                break;
+
+            case XFCE_NOTIFY_POS_BOTTOM_RIGHT:
+                pos_x = monitor_workarea->x + monitor_workarea->width - rectangle->x - rectangle->width;
+                pos_y = monitor_workarea->y + monitor_workarea->height - rectangle->y - rectangle->height;
+                break;
+
+            case XFCE_NOTIFY_POS_TOP_CENTER:
+            case XFCE_NOTIFY_POS_BOTTOM_CENTER: {
+                gint half_width = monitor_workarea->width / 2;
+                gboolean is_second_half = rectangle->x >= half_width;
+                gboolean is_ltr = gtk_widget_get_direction(GTK_WIDGET(window)) != GTK_TEXT_DIR_RTL;
+                if (!is_second_half) {
+                    // Place to the right (or left, if RTL) of the center.
+                    if (is_ltr) {
+                        pos_x = monitor_workarea->x + half_width + rectangle->x;
+                    } else {
+                        pos_x = monitor_workarea->x + monitor_workarea->width - rectangle->x - rectangle->width - half_width;
+                    }
+                } else {
+                    // The right (or left, if RTL) half of the screen is full, so
+                    // place on the left (or right, if RTL) side.
+                    if (is_ltr) {
+                        pos_x = monitor_workarea->x + monitor_workarea->width - rectangle->x - rectangle->width;
+                    } else {
+                        pos_x = monitor_workarea->x + rectangle->x;
+                    }
+                }
+
+                if (window->notify_location == XFCE_NOTIFY_POS_TOP_CENTER) {
+                    pos_y = monitor_workarea->y + rectangle->y;
+                } else {
+                    pos_y = monitor_workarea->y + monitor_workarea->height - rectangle->y - rectangle->height;
+                }
+
+                break;
+            }
+
+            default:
+                g_assert_not_reached();
+        }
+    } else
+#endif
+#ifdef ENABLE_WAYLAND
+    if (GDK_IS_WAYLAND_DISPLAY(gdk_display_get_default())) {
+        switch (window->notify_location) {
+            case XFCE_NOTIFY_POS_TOP_LEFT:
+            case XFCE_NOTIFY_POS_TOP_RIGHT:
+            case XFCE_NOTIFY_POS_BOTTOM_LEFT:
+            case XFCE_NOTIFY_POS_BOTTOM_RIGHT:
+                // Since the daemon is giving us positioning that's agnostic of
+                // the origin point for the notifications, and we've anchored
+                // the window in the appropriate origin corner, we can simply
+                // use the untranslated coordinates.
+                pos_x = rectangle->x;
+                break;
+
+            case XFCE_NOTIFY_POS_TOP_CENTER:
+            case XFCE_NOTIFY_POS_BOTTOM_CENTER: {
+                // We can use the untranslated y coordinate as above, but we
+                // need to move things around in the x direction so that we
+                // start in the center, and then "wrap" properly.  Compositors
+                // implementing wlr-layer-shell don't necessarily support
+                // negative margins, so instead of "non-anchoring" to the
+                // center, we have to calculate positions from one of the
+                // sides.
+                gboolean is_ltr = gtk_widget_get_direction(GTK_WIDGET(window)) != GTK_TEXT_DIR_RTL;
+                if (rectangle->x + rectangle->width < monitor_workarea->width / 2) {
+                    // First half of the monitor, so anchor to the left (or
+                    // right, for RTL) and then add half the monitor width so
+                    // it's placed in the center.
+                    gtk_layer_set_anchor(GTK_WINDOW(window),
+                                         is_ltr ? GTK_LAYER_SHELL_EDGE_LEFT : GTK_LAYER_SHELL_EDGE_RIGHT,
+                                         TRUE);
+                    pos_x = rectangle->x + monitor_workarea->width / 2;
+                } else {
+                    // Second half of the monitor, so anchor to the right (or
+                    // left, for RTL).  We don't need to modify the x
+                    // coordinate, as it's already set past the halfway point
+                    // on the monitor.
+                    gtk_layer_set_anchor(GTK_WINDOW(window),
+                                         is_ltr ? GTK_LAYER_SHELL_EDGE_RIGHT : GTK_LAYER_SHELL_EDGE_LEFT,
+                                         TRUE);
+                    pos_x = rectangle->x;
+                }
+                break;
+            }
+
+            default:
+                g_assert_not_reached();
+        }
+
+        pos_y = rectangle->y;
+    } else
+#endif
+    {
+        g_assert_not_reached();
+    }
+
+    window->geometry = *rectangle;
+    DBG("Translated position: (%d,%d)", pos_x, pos_y);
+    xfce_notify_window_move(window, pos_x, pos_y);
 }
 
 GdkRectangle *
